@@ -3,7 +3,7 @@ from django.forms.formsets import all_valid
 from django.forms.models import modelformset_factory
 from django.contrib.contenttypes.models import ContentType
 from django.contrib.admin import widgets, helpers
-from django.contrib.admin.util import unquote, flatten_fieldsets, get_deleted_objects, model_format_dict
+from django.contrib.admin.util import unquote, flatten_fieldsets, model_format_dict
 from django.contrib import messages
 from django.views.decorators.csrf import csrf_protect
 from django.core.exceptions import PermissionDenied, ValidationError
@@ -11,7 +11,11 @@ from django.core.paginator import Paginator
 from django.db import models, transaction, router
 from django.db.models.related import RelatedObject
 from django.db.models.fields import BLANK_CHOICE_DASH, FieldDoesNotExist
-from django.db.models.sql.constants import LOOKUP_SEP, QUERY_TERMS
+try:
+    from django.db.models.sql.constants import LOOKUP_SEP, QUERY_TERMS
+except ImportError:
+    from django.db.models.constants import LOOKUP_SEP
+    from django.db.models.sql.constants import QUERY_TERMS
 from django.http import Http404, HttpResponse, HttpResponseRedirect
 from django.shortcuts import get_object_or_404, render_to_response
 from django.utils.decorators import method_decorator
@@ -27,7 +31,7 @@ from django.utils.encoding import force_unicode
 from django.forms.forms import pretty_name
 
 from mongoengine.fields import (DateTimeField, URLField, IntField, ListField, EmbeddedDocumentField, 
-                                ReferenceField, StringField, FileField)
+                                ReferenceField, StringField, FileField, ImageField)
 
 from mongodbforms.documentoptions import DocumentMetaWrapper
 from mongoadmin import mongohelpers
@@ -35,7 +39,7 @@ from mongoadmin.util import RelationWrapper
 
 from mongodbforms.documents import (documentform_factory, DocumentForm, 
                                   inlineformset_factory, BaseInlineDocumentFormSet)
-from mongodbforms import MongoDefaultFormFieldGenerator, init_document_options, save_instance
+from mongodbforms import MongoDefaultFormFieldGenerator, save_instance
 
 HORIZONTAL, VERTICAL = 1, 2
 # returns the <ul> class for a given radio_admin field
@@ -55,8 +59,8 @@ FORMFIELD_FOR_DBFIELD_DEFAULTS = {
     #models.TimeField:       {'widget': widgets.AdminTimeWidget},
     URLField:       {'widget': widgets.AdminURLFieldWidget},
     IntField:       {'widget': widgets.AdminIntegerFieldWidget},
-    #models.ImageField:      {'widget': widgets.AdminFileWidget},
-    FileField:       {'widget': widgets.AdminFileWidget},
+    ImageField:     {'widget': widgets.AdminFileWidget},
+    FileField:      {'widget': widgets.AdminFileWidget},
 }
 
 csrf_protect_m = method_decorator(csrf_protect)
@@ -116,6 +120,15 @@ class BaseDocumentAdmin(object):
         overrides = FORMFIELD_FOR_DBFIELD_DEFAULTS.copy()
         overrides.update(self.formfield_overrides)
         self.formfield_overrides = overrides
+        
+    def _get_formfield(self, db_field, **kwargs):
+        """Return overridden formfield if exists, otherwise default formfield"""
+        # If we've got overrides for the formfield defined, use 'em. **kwargs
+        # passed to formfield_for_dbfield override the defaults.
+        for klass in db_field.__class__.mro():
+            if klass in self.formfield_overrides:
+                kwargs = dict(self.formfield_overrides[klass], **kwargs)
+        return formfield(db_field, **kwargs)        
 
     def formfield_for_dbfield(self, db_field, **kwargs):
         """
@@ -140,7 +153,7 @@ class BaseDocumentAdmin(object):
             # extra HTML -- the "add other" interface -- to the end of the
             # rendered output. formfield can be None if it came from a
             # OneToOneField with parent_link=True or a M2M intermediary.
-            form_field = formfield(db_field, **kwargs)
+            form_field = self._get_formfield(db_field, **kwargs)
             if db_field.name not in self.raw_id_fields:
                 related_modeladmin = self.admin_site._registry.get(db_field.document_type)
                 can_add_related = bool(related_modeladmin and
@@ -155,17 +168,10 @@ class BaseDocumentAdmin(object):
                 kwargs = dict({'widget': widgets.AdminTextareaWidget}, **kwargs)
             else:
                 kwargs = dict({'widget': widgets.AdminTextInputWidget}, **kwargs)
-            return formfield(db_field, **kwargs)
+            return self._get_formfield(db_field, **kwargs)
         
-        # If we've got overrides for the formfield defined, use 'em. **kwargs
-        # passed to formfield_for_dbfield override the defaults.
-        for klass in db_field.__class__.mro():
-            if klass in self.formfield_overrides:
-                kwargs = dict(self.formfield_overrides[klass], **kwargs)
-                return formfield(db_field, **kwargs)
-            
         # For any other type of field, just call its formfield() method.
-        return formfield(db_field, **kwargs)
+        return self._get_formfield(db_field, **kwargs)
 
     def formfield_for_choice_field(self, db_field, request=None, **kwargs):
         """
@@ -279,6 +285,8 @@ class DocumentAdmin(BaseDocumentAdmin):
     list_display_links = ()
     list_filter = ()
     list_select_related = False
+    # see __init__ for django < 1.4
+    list_max_show_all = 200
     list_per_page = 100
     list_editable = ()
     search_fields = ()
@@ -328,14 +336,21 @@ class DocumentAdmin(BaseDocumentAdmin):
         # instances derived from DocumentAdmin. 
         self.exclude = list(self.exclude)
         self.get_inline_instances()
-        
-        if 'action_checkbox' not in self.list_display and self.actions is not None:
-            self.list_display = ['action_checkbox'] +  list(self.list_display)
-        if not self.list_display_links:
-            for name in self.list_display:
-                if name != 'action_checkbox':
-                    self.list_display_links = [name]
-                    break
+                
+        # If someone patched their MAX_SHOW_ALL_ALLOWED in django 1.3, we 
+        # get the value here and proceed as normal.
+        try:
+            from django.contrib.admin.views.main import MAX_SHOW_ALL_ALLOWED
+            self.list_max_show_all = MAX_SHOW_ALL_ALLOWED
+        except ImportError:
+            pass
+
+    @property
+    def content_type(self):
+        name = app_label, model = self.opts.app_label, self.opts.module_name
+        name = '%s %s'%name
+        ret = ContentType.objects.get_or_create(name=name, app_label=app_label, model=model)
+        return ret[0]
 
     def get_inline_instances(self):
         for f in self.document._fields.itervalues():
@@ -352,7 +367,7 @@ class DocumentAdmin(BaseDocumentAdmin):
             else:
                 # For some reason we found an embedded field were either
                 # the field attribute or the field's document type is None.
-                # This shouldn't happen, but appearently does happen:
+                # This shouldn't happen, but apparently does happen:
                 # https://github.com/jschrewe/django-mongoadmin/issues/4
                 # The solution for now is to ignore that field entirely.
                 continue
@@ -565,7 +580,7 @@ class DocumentAdmin(BaseDocumentAdmin):
         from django.contrib.admin.models import LogEntry, ADDITION
         LogEntry.objects.log_action(
             user_id         = request.user.pk,
-            content_type_id = None,
+            content_type_id = self.content_type.id,
             object_id       = object.pk,
             object_repr     = force_unicode(object),
             action_flag     = ADDITION
@@ -580,7 +595,7 @@ class DocumentAdmin(BaseDocumentAdmin):
         from django.contrib.admin.models import LogEntry, CHANGE
         LogEntry.objects.log_action(
             user_id         = request.user.pk,
-            content_type_id = None,
+            content_type_id = self.content_type.id,
             object_id       = object.pk,
             object_repr     = force_unicode(object),
             action_flag     = CHANGE,
@@ -597,7 +612,7 @@ class DocumentAdmin(BaseDocumentAdmin):
         from django.contrib.admin.models import LogEntry, DELETION
         LogEntry.objects.log_action(
             user_id         = request.user.id,
-            content_type_id = None,
+            content_type_id = self.content_type.id,
             object_id       = object.pk,
             object_repr     = object_repr,
             action_flag     = DELETION
@@ -643,7 +658,6 @@ class DocumentAdmin(BaseDocumentAdmin):
 
         # Convert the actions into a SortedDict keyed by name
         # and sorted by description.
-        actions.sort(key=lambda k: k[2].lower())
         actions = SortedDict([
             (name, (func, name, desc))
             for func, name, desc in actions
@@ -692,6 +706,31 @@ class DocumentAdmin(BaseDocumentAdmin):
             description = capfirst(action.replace('_', ' '))
         return func, action, description
 
+    def get_list_display(self, request):
+        """
+        Return a sequence containing the fields to be displayed on the
+        changelist.
+        """
+        return self.list_display
+
+    def get_list_display_links(self, request, list_display):
+        """
+        Return a sequence containing the fields to be displayed as links
+        on the changelist. The list_display parameter is the list of fields
+        returned by get_list_display().
+        """
+        if self.list_display_links or not list_display:
+            return self.list_display_links
+        else:
+            # Use only the first item in list_display as link
+            return list(list_display)[:1]
+        
+    def get_ordering(self, request):
+        """
+        Hook for specifying field ordering.
+        """
+        return self.ordering or ()  # otherwise we might try to *None, which is bad ;)
+
     def construct_change_message(self, request, form, formsets):
         """
         Construct a change message from a changed object.
@@ -908,16 +947,7 @@ class DocumentAdmin(BaseDocumentAdmin):
                 # Perform the action only on the selected objects
                 queryset = queryset.filter(pk__in=selected)
 
-            qs = []
-            for m in queryset:
-                if isinstance(m._meta, DocumentMetaWrapper):
-                    m._meta.document = m
-                    m._meta.init_pk()
-                else:
-                    m._meta = DocumentMetaWrapper(m)
-                qs.append(m)
-
-            response = func(self, request, qs)
+            response = func(self, request, queryset)
 
             # Actions may return an HttpResponse, which will be used as the
             # response from the POST. If not, we'll be a good little HTTP
@@ -1151,22 +1181,22 @@ class DocumentAdmin(BaseDocumentAdmin):
         if not self.has_change_permission(request, None):
             raise PermissionDenied
 
+        list_display = self.get_list_display(request)
+        list_display_links = self.get_list_display_links(request, list_display)
+
         # Check actions to see if any are available on this changelist
         actions = self.get_actions(request)
-
-        # Remove action checkboxes if there aren't any actions available.
-        list_display = list(self.list_display)
-        if not actions:
-            try:
-                list_display.remove('action_checkbox')
-            except ValueError:
-                pass
+        if actions:
+            # Add the action checkboxes if there are any actions available.
+            list_display = ['action_checkbox'] +  list(list_display)
 
         ChangeList = self.get_changelist(request)
         try:
-            cl = ChangeList(request, self.model, list_display, self.list_display_links,
-                self.list_filter, self.date_hierarchy, self.search_fields,
-                self.list_select_related, self.list_per_page, self.list_editable, self)
+            cl = ChangeList(request, self.model, list_display,
+                list_display_links, self.list_filter, self.date_hierarchy,
+                self.search_fields, self.list_select_related,
+                self.list_per_page, self.list_max_show_all, self.list_editable,
+                self) 
         except IncorrectLookupParameters:
             # Wacky lookup parameters were given, so redirect to the main
             # changelist page, without parameters, and pass an 'invalid=1'
@@ -1361,10 +1391,16 @@ class DocumentAdmin(BaseDocumentAdmin):
         app_label = opts.app_label
         action_list = LogEntry.objects.filter(
             object_id = object_id,
-            content_type__id__exact = ContentType.objects.get_for_model(model).id
+            content_type__id__exact = self.content_type.id
         ).select_related().order_by('action_time')
         # If no history was found, see whether this object even exists.
-        obj = get_object_or_404(model, pk=unquote(object_id))
+
+        from mongoengine.base import ValidationError
+        try:
+            obj = model.objects.get(id=object_id)
+        except ValidationError, e:
+            raise Http404('No %s matches the given query.' % model._meta.object_name)
+
         context = {
             'title': _('Change history: %s') % force_unicode(obj),
             'action_list': action_list,
@@ -1372,6 +1408,7 @@ class DocumentAdmin(BaseDocumentAdmin):
             'object': obj,
             'root_path': self.admin_site.root_path,
             'app_label': app_label,
+            'opts': opts,
         }
         context.update(extra_context or {})
         context_instance = template.RequestContext(request, current_app=self.admin_site.name)
